@@ -3,13 +3,10 @@ Utilities for creating typed Python dataclasses for storing hyperparameters.
 """
 from typing import TYPE_CHECKING, Any, Dict, Mapping
 
+from scalarstop._attr_dict import AttrDict
 from scalarstop._naming import hash_id
 from scalarstop.dataclasses import asdict, is_dataclass
-from scalarstop.exceptions import (
-    WrongHyperparamsKeys,
-    WrongHyperparamsType,
-    YouForgotTheHyperparams,
-)
+from scalarstop.exceptions import WrongHyperparamsType, YouForgotTheHyperparams
 
 if TYPE_CHECKING:
     from dataclasses import dataclass
@@ -20,7 +17,7 @@ else:
 def enforce_dict(hyperparams: Any) -> Dict[Any, Any]:
     """Convert the input into a dictionary, whether it is a dataclass or not."""
     if hyperparams is None:
-        return dict()
+        return {}
     if isinstance(hyperparams, dict):
         return hyperparams
     try:
@@ -40,25 +37,33 @@ def init_hyperparams(*, class_name: str, hyperparams, hyperparams_class) -> Any:
     """
     if isinstance(hyperparams_class, type) and is_dataclass(hyperparams_class):
         if hyperparams is None:
-            try:
-                return hyperparams_class()
-            except (TypeError, ValueError, SyntaxError) as exc:
-                raise WrongHyperparamsKeys(
-                    hyperparams=hyperparams, hyperparams_class=hyperparams_class
-                ) from exc
+            return hyperparams_class()
         if isinstance(hyperparams, hyperparams_class):
             return hyperparams
         if isinstance(hyperparams, Mapping):
-            try:
-                return hyperparams_class(**hyperparams)
-            except (TypeError, ValueError, SyntaxError) as exc:
-                raise WrongHyperparamsKeys(
-                    hyperparams=hyperparams, hyperparams_class=hyperparams_class
-                ) from exc
+            return hyperparams_class(**hyperparams)
         raise WrongHyperparamsType(hyperparams=hyperparams, class_name=class_name)
     raise YouForgotTheHyperparams(class_name=class_name)
 
 
+def _flatten_hyperparams(hp_dict):
+    parent_hp_dict = hp_dict.pop("parent", None)
+    if parent_hp_dict:
+        return {
+            **_flatten_hyperparams(parent_hp_dict["hyperparams"]),
+            **hp_dict,
+        }
+    return hp_dict
+
+
+def flatten_hyperparams(hyperparams: Any) -> Dict[str, Any]:
+    """
+    Recursively flatten the hyperparams embedded
+    in a :py:class:`NestedHyperparamsType` instance.
+    """
+    return AttrDict(**_flatten_hyperparams(enforce_dict(hyperparams)))
+
+
 @dataclass  # pylint: disable=used-before-assignment
 class HyperparamsType:
     """
